using System.Security.Cryptography;
using System.Text;
using FitConnect.Models.Api.Metadata;

namespace FitConnect.Models;

public class Attachment {
    public Attachment(string id, Api.Metadata.Attachment metadata, byte[] content,
        string attachmentAuthentication) {
        Id = id;
        Filename = metadata.Filename;
        Content = content;
        AttachmentAuthentication = attachmentAuthentication;
        MimeType = Path.GetExtension(metadata.Filename) switch {
            "pdf" => "application/pdf",
            "xml" => "application/xml",
            _ => ""
        };
    }

    /// <summary>
    ///     Attachment for a submission.
    /// </summary>
    /// <param name="filename">File name to load</param>
    /// <param name="description">Description of the attachment</param>
    public Attachment(string filename, string description, byte[] content) {
        var file = new FileInfo(filename);
        var extension = file.Extension[1..].ToLower();
        Filename = filename;
        MimeType = "application/" + extension;
        Description = description;

        Content = content;

        Signature = new AttachmentSignature {
            DetachedSignature = true,
            SignatureFormat = extension switch {
                "pdf" => SignatureFormat.Pdf,
                "xml" => SignatureFormat.Xml,
                "json" => SignatureFormat.Json,
                _ => throw new ArgumentOutOfRangeException()
            }
        };
    }

    /// <summary>
    ///     Attachment for a submission.
    /// </summary>
    /// <param name="filename">File name to load</param>
    /// <param name="description">Description of the attachment</param>
    public Attachment(string filename, string description) : this(filename, description,
        File.ReadAllBytes(filename)) {
    }


    public string Id { get; init; } = Guid.NewGuid().ToString();

    public byte[]? Content { get; init; }
    public string? AttachmentAuthentication { get; }

    public string? Hash => CalculateHash();

    public string? Filename { get; init; }

    public string? Purpose { get; init; } = "attachment";

    public string MimeType { get; init; }

    public string? Description { get; init; }

    public AttachmentSignature? Signature { get; }

    private string CalculateHash() {
        return ByteToHexString(SHA512.Create().ComputeHash(Content ?? Array.Empty<byte>()));
    }

    private static string ByteToHexString(IEnumerable<byte> data) {
        var sb = new StringBuilder();
        foreach (var b in data) sb.Append(b.ToString("x2"));

        return sb.ToString();
    }
}
